/*
 * casocklib - An asynchronous communication library for C++
 * ---------------------------------------------------------
 * Copyright (C) 2010 Leandro Costa
 *
 * This file is part of casocklib.
 *
 * casocklib is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * casocklib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with casocklib. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file casock/rpc/asio/protobuf/client/RPCClientProxy.cc
 * \brief [brief description]
 * \author Leandro Costa
 * \date 2010
 *
 * $LastChangedDate$
 * $LastChangedBy$
 * $Revision$
 */

#include "casock/rpc/asio/protobuf/client/RPCClientProxy.h"

#include <boost/bind.hpp>
#include <google/protobuf/descriptor.h>

#include "casock/util/Logger.h"
#include "casock/rpc/protobuf/api/rpc.pb.h"
#include "casock/rpc/protobuf/client/RPCCall.h"
#include "casock/rpc/protobuf/client/RPCCallQueue.h"
#include "casock/rpc/protobuf/client/RPCCallHandler.h"
#include "casock/rpc/asio/protobuf/client/RPCSocketClient.h"
#include "casock/rpc/asio/protobuf/client/RPCSocketClientFactory.h"
#include "casock/rpc/asio/protobuf/client/RPCClientCommunicator.h"

#include "casock/rpc/asio/protobuf/client/RPCClientCommunicatorImpl.h"
#include "casock/proactor/asio/base/AsyncProcessor.h"
#include "casock/util/Lock.h"
#include "casock/proactor/asio/base/AsyncProcessor.h"

namespace casock {
  namespace rpc {
    namespace asio {
      namespace protobuf {
        namespace client {
          RPCClientProxy::RPCClientProxy (const RPCSocketClientFactory& rSocketClientFactory, const RPCCallHandlerFactory& rCallHandlerFactory)
            : casock::rpc::protobuf::client::RPCClientProxy (rCallHandlerFactory)
          {
            LOGMSG (HIGH_LEVEL, "%s\n", __PRETTY_FUNCTION__);
            //Xuxi for reconnect, RPCSocketClientFactoryImpl instance in stack, but need reused, so...clone...
            this->mpSocketClientFactory = rSocketClientFactory.clone();
            mpSocketClient = NULL;
            mpCommunicator = NULL;
            init(); //maybe needn't init here...
//            mpSocketClient = rSocketClientFactory.buildRPCSocketClient ();
//            mpCommunicator = mpSocketClient->buildCommunicator (*mpCallHash, *mpCallQueue);
//            mpCommunicator->startReceivingResponses ();
          }

          //add by Xuxi for reconnect
          void RPCClientProxy::init(){
        	  LOGMSG (HIGH_LEVEL, "**2**%s this=[%p]\n", __PRETTY_FUNCTION__, this);
        	  if(mpSocketClient!=NULL){ delete mpSocketClient;mpSocketClient=NULL;}
              mpSocketClient = mpSocketClientFactory->buildRPCSocketClient ();

        	  if(mpCommunicator != NULL) {delete mpCommunicator;mpCommunicator=NULL;}
              mpCommunicator = mpSocketClient->buildCommunicator (*mpCallHash, *mpCallQueue);

              mpCommunicator->startReceivingResponses ();
              this->initSuccess=true;
          }

          RPCClientProxy::~RPCClientProxy ()
          {
            delete mpSocketClient;
            delete mpCommunicator;
            delete mpSocketClientFactory;
          }

          void RPCClientProxy::sendRpcRequest (const casock::rpc::protobuf::api::RpcRequest& request, casock::rpc::protobuf::client::RPCCall* pCall)
          {
            LOGMSG (HIGH_LEVEL, "RPCClientProxy::%s ()\n", __FUNCTION__);
            //add by Xuxi for reconnect begin
            RPCClientCommunicatorImpl* pCommunicator = (RPCClientCommunicatorImpl*)mpCommunicator;
			if(! this->initSuccess || mpCommunicator==NULL ||  pCommunicator->isBreaked() ){
			    try
			    {
					this->init();

			    }
			    catch (std::exception& e)
			    {
			      LOGMSG (NO_DEBUG, "RPCClientProxy::%s () - init error: [%s]\n", __FUNCTION__, e.what ());
	        	  if(mpSocketClient!=NULL){ delete mpSocketClient;mpSocketClient=NULL;}
	        	  if(mpCommunicator != NULL) {delete mpCommunicator;mpCommunicator=NULL;}
			      this->initSuccess=false;
			      throw e;
			    }
				::casock::util::Lock& lock = ::casock::proactor::asio::base::AsyncProcessor::getInstance()->getLock();
				lock.get ();
				lock.signal();
				lock.release ();
            }
			registerRPCCall (request.id (), pCall);
            //add by Xuxi for reconnect end

            mpCommunicator->sendRequest (request, ::boost::bind (&RPCClientProxy::onSentRequest, this, _1, request.id (), pCall));
          }

					void RPCClientProxy::onSentRequest (const ::asio::error_code& error, const uint32 id, casock::rpc::protobuf::client::RPCCall* pCall)
					{
            LOGMSG (HIGH_LEVEL, "RPCClientProxy::%s ()\n", __FUNCTION__);

						if (! error)
						{
							//comment by Xuxi for timeout error, move to sendRpcRequest, because it might later than receive response.
							//registerRPCCall (id, pCall);
						}
						else
						{
							LOGMSG (HIGH_LEVEL, "***%s--Impossible to send request[%d]:[%p] because:[%s]h\n ", __FUNCTION__, id, pCall,error.message ());
							pCall->controller ()->SetFailed ("Impossible to send request [" + error.message () + "]");
							mpCallQueue->push (pCall);
						}
					}

          void RPCClientProxy::close ()
          {
            mpSocketClient->close ();
          }
        }
      }
    }
  }
}
