/*
 * SynRPCClientProxy.cpp
 *
 *  Created on: Jun 10, 2011
 *      Author: root
 */

#include "casock/rpc/synio/protobuf/client/SynRPCClientProxy.h"
#include <cerrno>
#include <stdexcept>
#include <google/protobuf/descriptor.h>
#include "casock/util/Logger.h"
#include "casock/rpc/protobuf/api/rpc.pb.h"

namespace casock {

namespace rpc {

namespace synio {

namespace protobuf {

namespace client {
const RPCRequestBuilder* SynRPCClientProxy::smpRequestBuilder = (new RPCRequestBuilder());

SynRPCClientProxy::SynRPCClientProxy(const string& ip, const int& port) :
	mIp(ip), mPort(port),mSocketFd(0),
			initSuccess(false) {
	LOGMSG(HIGH_LEVEL, "%s\n", __PRETTY_FUNCTION__);
}

SynRPCClientProxy::~SynRPCClientProxy() {
	LOGMSG(HIGH_LEVEL, "SynRPCClientProxy::%s ()\n", __FUNCTION__);
	//delete mpRequestBuilder;
	close();
}

void SynRPCClientProxy::init() {
	close();
	this->initSuccess = false;
	if ((-1 == (this->mSocketFd = socket(AF_INET, SOCK_STREAM, 0)))) {
		LOGMSG(HIGH_LEVEL, "SynRPCClientProxy::%s () create socket failed[%s]\n",
				__FUNCTION__, strerror(errno));
		throw std::runtime_error(string("create socket: ")+strerror(errno));
		return; //never go here for exception throws
	}

	struct sockaddr_in ServerAddr;
	memset(&ServerAddr, 0, sizeof(struct sockaddr_in));
	ServerAddr.sin_family = AF_INET;

	ServerAddr.sin_addr.s_addr = inet_addr(this->mIp.c_str());
	ServerAddr.sin_port = htons(this->mPort);
	unsigned int uLen = sizeof(struct sockaddr_in);
	if (-1 == (connect(this->mSocketFd, (struct sockaddr *) &ServerAddr, uLen))) {
		LOGMSG(HIGH_LEVEL,
				"SynRPCClientProxy::%s () connect to TCP server failed[%s]\n",
				__FUNCTION__,strerror(errno));
		throw std::runtime_error("connect to TCP server failed");
		return;
	}
	this->initSuccess = true;
}

void SynRPCClientProxy::sendRpcRequest(const google::protobuf::Message* message) {
	LOGMSG(HIGH_LEVEL, "SynRPCClientProxy::%s ()\n", __FUNCTION__);
	size_t size = message->ByteSize();
	std::stringstream* pBuffer = new std::stringstream(std::ios_base::in
			| std::ios_base::out | std::ios_base::binary);
	pBuffer->write(reinterpret_cast<const char*> (&size), sizeof(size_t));
	message->SerializeToOstream(pBuffer);
	bool reconnected=false;
	int nCount = 0;
	while (nCount < size) {
		int nSendLen = -1;
		if ((nSendLen = send(this->mSocketFd, pBuffer->str().c_str(), size
				+ sizeof(size_t) - nCount, NULL)) < 0) { // send error

			if(!reconnected){ //try reconnect and send again.
				LOGMSG(HIGH_LEVEL,
									"SynRPCClientProxy::%s () reconnect to TCP server %s:%d for error:[%d][%s]\n",
									__FUNCTION__, this->mIp, this->mPort, errno, strerror(errno));
				this->close();
				this->init();
				reconnected = true;
				continue;
			}
			delete pBuffer;
			LOGMSG(HIGH_LEVEL,
					"SynRPCClientProxy::%s () send to TCP server failed\n",
					__FUNCTION__);
			throw (std::runtime_error("send to TCP server failed"));
			return;
		}
		nCount += nSendLen;
	}
	delete pBuffer;
}
/**
 * skip buffer data and check if socket is closed;
 */
bool SynRPCClientProxy::isBroken(){

	size_t size;
	int iRead;
	//int iRtn = recv(this->mSocketFd, (char *) &size, sizeof(size_t), MSG_DONTWAIT | MSG_PEEK);
	while((iRead = recv(this->mSocketFd, (char *) &size, sizeof(size_t), MSG_DONTWAIT )) > 0);
	if(iRead== 0){ // return -1 means has no data read, but connection still alive
		return true;
	}
	else { //if(iRtn < 0)
		return false;
	}
}

void SynRPCClientProxy::resvRpcResponse(google::protobuf::Message* pResponse) {
	LOGMSG(HIGH_LEVEL, "SynRPCClientProxy::%s ()\n", __FUNCTION__);
	//set timeout for socket
	setsockopt(this->mSocketFd, SOL_SOCKET, SO_RCVTIMEO, &this->timeout(),
			sizeof(struct timeval));
	size_t size;
	int iRtn = recv(this->mSocketFd, (char *) &size, sizeof(size_t), 0);
	if (iRtn < 0) {
		LOGMSG(HIGH_LEVEL,
				"SynRPCClientProxy::%s () recv from TCP server failed,errno:[%d]=[%s]\n",
				__FUNCTION__,errno,strerror(errno));
		if(errno != EAGAIN){
			this->close();
			this->initSuccess = false;
		}
		throw (std::runtime_error(string("recv from TCP server error:")+strerror(errno)));
		return;
	} else if (iRtn == 0) { //remote closed
		LOGMSG(HIGH_LEVEL,
				"SynRPCClientProxy::%s () recv size from  server but remote closed [%s]\n",
				__FUNCTION__,strerror(errno));
		this->close();
		this->initSuccess = false;
		throw (std::runtime_error("recv size from  server but remote closed "));
		return;
	}
	char* pBuffer = new char[size];

	int nCount = 0;
	while (nCount < size) {
		int nLen = -1;
		if ((nLen = recv(this->mSocketFd, pBuffer, size - nCount, nCount)) < 0) { //time out
			delete[] pBuffer;
			this->close();
			this->initSuccess = false;
			LOGMSG(HIGH_LEVEL,
					"SynRPCClientProxy::%s () recv from TCP server failed[%s]\n",
					__FUNCTION__,strerror(errno));
			if(errno != EAGAIN){
				this->close();
				this->initSuccess = false;
			}
			throw (std::runtime_error(string("recv from TCP server error:")+strerror(errno)));
			return;
		} else if (nLen == 0) { //remote closed
			LOGMSG(HIGH_LEVEL,
					"SynRPCClientProxy::%s () recv size:0 [%s]\n",
					__FUNCTION__,strerror(errno));
			this->close();
			this->initSuccess = false;
			throw (std::runtime_error("recv size from TCP server timeout"));
			return;
		}
		nCount += nLen;
	}

	std::stringstream ss(std::string(pBuffer, size));
	pResponse->ParseFromIstream(&ss);
	delete[] pBuffer;
	return;
}

void SynRPCClientProxy::close() {
	if(this->mSocketFd){
		::close(this->mSocketFd);
		this->mSocketFd = 0;
	}
}

void SynRPCClientProxy::CallMethod(
		const google::protobuf::MethodDescriptor* method,
		google::protobuf::RpcController* controller,
		const google::protobuf::Message* request,
		google::protobuf::Message* response, google::protobuf::Closure* done) {
	LOGMSG(HIGH_LEVEL, "RPCClientProxy::%s ()\n", __FUNCTION__);
	this->lock.get();
	if (!this->initSuccess || this->isBroken()) {
		try {
			this->init();
		} catch (std::runtime_error &ex) {
			if (controller) {
				controller->SetFailed(ex.what());
			}
			this->lock.release();
			throw ex;
		}
	}
	casock::rpc::protobuf::api::RpcRequest* pRpcRequest =
			smpRequestBuilder->buildRpcRequest(method, request);
	try {
		sendRpcRequest(pRpcRequest);
	} catch (std::runtime_error &ex) {
		if (controller) {
			controller->SetFailed(ex.what());
		}
		delete pRpcRequest;
		this->lock.release();
		throw ex;
	}
	casock::rpc::protobuf::api::RpcResponse* pRpcResponse =
			new casock::rpc::protobuf::api::RpcResponse();
	try {
		resvRpcResponse(pRpcResponse);
		while(pRpcRequest->id() != pRpcResponse->id()){ //recv till the id found.
			LOGMSG(HIGH_LEVEL, "RPCClientProxy::%s () received a not correct answer, try again\n", __FUNCTION__);
			resvRpcResponse(pRpcResponse);
		}
		response->ParseFromString(pRpcResponse ->response());
	} catch (std::runtime_error &ex) {
		if (controller) {
			controller->SetFailed(ex.what());
		}
		delete pRpcResponse;
		this->lock.release();
		throw ex;
	}
	if(done){
		done->Run();
	}
	delete pRpcResponse;
	delete pRpcRequest;
	this->lock.release();
}

RPCCallController* SynRPCClientProxy::buildRPCCallController () const
{
  LOGMSG (LOW_LEVEL, "%s\n - timeOut[%d] seconds", __PRETTY_FUNCTION__,this->timeout().tv_sec);
  RPCCallController* pController = new RPCCallController ();
  pController->setTimeout (this->timeout ());

  return pController;
}

}

}

}

}

}
